import { Node } from "./node.js";
interface EdgeOptions {
    id: string | number;
    sourceId: string | number;
    targetId: string | number;
    directed?: boolean;
    data: Record<string, unknown>;
}
export declare class Edge {
    /** Unique uuid of the edge. */
    id: string | number;
    /** ID of the source node. */
    private _sourceId;
    /** ID of the target node. */
    private _targetId;
    /** Whether the edge is directed or not. */
    private _directed;
    /** Origin data reference of the edge. */
    private _data;
    /** Check the type of the object when picking engine gets it. */
    readonly isEdge = true;
    /** Nodes at either end of this edge. */
    private readonly _connectedNodes;
    /** Edge state. */
    state: string;
    /**
     * The constructor
     * @param  {String|Number} options.id - the unique ID of the edge
     * @param  {String|Number} options.sourceId - the ID of the source node
     * @param  {String|Number} options.targetId - the ID of the target node
     * @param  {Boolean} options.directed - whether the edge is directed or not
     * @param  {Record<string, unknown>} options.data - origin data reference
     */
    constructor({ id, sourceId, targetId, data, directed }: EdgeOptions);
    /**
     * Return the ID of the edge
     * @return {String|Number} - the ID of the edge.
     */
    getId(): string | number;
    /**
     * Return whether the edge is directed or not.
     * @return {Boolean} true if the edge is directed.
     */
    isDirected(): boolean;
    /**
     * Get the ID of the source node.
     * @return {String|Number} the ID of the source node.
     */
    getSourceNodeId(): string | number;
    /**
     * Get the ID of the target node.
     * @return {String|Number} the ID of the target node.
     */
    getTargetNodeId(): string | number;
    /**
     * Return of the value of the selected property key.
     * @param  {String} key - property key.
     * @return {Any} - the value of the property.
     */
    getPropertyValue(key: string): unknown;
    /**
     * Set the origin data as a reference.
     * @param {Object} data - the origin data.
     */
    setData(data: Record<string, unknown>): void;
    /**
     * Update a data property.
     * @param {String} key - the key of the property
     * @param {Any} value - the value of the property.
     */
    setDataProperty(key: string, value: unknown): void;
    /**
     * Set edge state
     * @param {String} state - one of EDGE_STATE
     */
    setState(state: string): void;
    /**
     * Get edge state
     * @returns {string} state - one of EDGE_STATE
     */
    getState(): string;
    addNode(node: Node): void;
    removeNode(node: Node): void;
    getConnectedNodes(): Node[];
}
export {};
//# sourceMappingURL=edge.d.ts.map